@@ -47,26 +47,14 @@
 static QTime sChrono;
 #endif
 
-typedef void (DBusMenuImporter::*DBusMenuImporterMethod)(int, QDBusPendingCallWatcher*);
-
 static const char *DBUSMENU_INTERFACE = "com.canonical.dbusmenu";
 
 static const int ABOUT_TO_SHOW_TIMEOUT = 3000;
 static const int REFRESH_TIMEOUT = 4000;
 
 static const char *DBUSMENU_PROPERTY_ID = "_dbusmenu_id";
-static const char *DBUSMENU_PROPERTY_ICON = "_dbusmenu_icon";
-
-struct Task
-{
-    Task()
-    : m_id(0)
-    , m_method(0)
-    {}
-
-    int m_id;
-    DBusMenuImporterMethod m_method;
-};
+static const char *DBUSMENU_PROPERTY_ICON_NAME = "_dbusmenu_icon_name";
+static const char *DBUSMENU_PROPERTY_ICON_DATA_HASH = "_dbusmenu_icon_data_hash";
 
 static QAction *createKdeTitle(QAction *action, QWidget *parent)
 {
@@ -91,7 +79,6 @@ public:
 
     QDBusAbstractInterface *m_interface;
     QMenu *m_menu;
-    QMap<QDBusPendingCallWatcher *, Task> m_taskForWatcher;
     typedef QMap<int, QPointer<QAction> > ActionForId;
     ActionForId m_actionForId;
     QSignalMapper m_mapper;
@@ -110,13 +97,9 @@ public:
         #endif
         QDBusPendingCall call = m_interface->asyncCall("GetLayout", id, 1, QStringList());
         QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(call, q);
+        watcher->setProperty(DBUSMENU_PROPERTY_ID, id);
         QObject::connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)),
-            q, SLOT(dispatch(QDBusPendingCallWatcher*)));
-
-        Task task;
-        task.m_id = id;
-        task.m_method = &DBusMenuImporter::GetLayoutCallback;
-        m_taskForWatcher.insert(watcher, task);
+            q, SLOT(slotGetLayoutFinished(QDBusPendingCallWatcher*)));
 
         return watcher;
     }
@@ -126,6 +109,8 @@ public:
         QMenu *menu = q->createMenu(parent);
         QObject::connect(menu, SIGNAL(aboutToShow()),
             q, SLOT(slotMenuAboutToShow()));
+        QObject::connect(menu, SIGNAL(aboutToHide()),
+            q, SLOT(slotMenuAboutToHide()));
         return menu;
     }
 
@@ -197,7 +182,9 @@ public:
         } else if (key == "toggle-state") {
             updateActionChecked(action, value);
         } else if (key == "icon-name") {
-            updateActionIcon(action, value);
+            updateActionIconByName(action, value);
+        } else if (key == "icon-data") {
+            updateActionIconByData(action, value);
         } else if (key == "visible") {
             updateActionVisible(action, value);
         } else if (key == "shortcut") {
@@ -225,14 +212,14 @@ public:
         }
     }
 
-    void updateActionIcon(QAction *action, const QVariant &value)
+    void updateActionIconByName(QAction *action, const QVariant &value)
     {
         QString iconName = value.toString();
-        QString previous = action->property(DBUSMENU_PROPERTY_ICON).toString();
+        QString previous = action->property(DBUSMENU_PROPERTY_ICON_NAME).toString();
         if (previous == iconName) {
             return;
         }
-        action->setProperty(DBUSMENU_PROPERTY_ICON, iconName);
+        action->setProperty(DBUSMENU_PROPERTY_ICON_NAME, iconName);
         if (iconName.isEmpty()) {
             action->setIcon(QIcon());
             return;
@@ -240,6 +227,24 @@ public:
         action->setIcon(q->iconForName(iconName));
     }
 
+    void updateActionIconByData(QAction *action, const QVariant &value)
+    {
+        QByteArray data = value.toByteArray();
+        uint dataHash = qHash(data);
+        uint previousDataHash = action->property(DBUSMENU_PROPERTY_ICON_DATA_HASH).toUInt();
+        if (previousDataHash == dataHash) {
+            return;
+        }
+        action->setProperty(DBUSMENU_PROPERTY_ICON_DATA_HASH, dataHash);
+        QPixmap pix;
+        if (!pix.loadFromData(data)) {
+            DMWARNING << "Failed to decode icon-data property for action" << action->text();
+            action->setIcon(QIcon());
+            return;
+        }
+        action->setIcon(QIcon(pix));
+    }
+
     void updateActionVisible(QAction *action, const QVariant &value)
     {
         action->setVisible(value.isValid() ? value.toBool() : true);
@@ -260,11 +265,20 @@ public:
             return q->menu();
         }
         QAction *action = m_actionForId.value(id);
-        DMRETURN_VALUE_IF_FAIL(action, 0);
+        if (!action) {
+            return 0;
+        }
         return action->menu();
     }
 
     void slotItemsPropertiesUpdated(const DBusMenuItemList &updatedList, const DBusMenuItemKeysList &removedList);
+
+    void sendEvent(int id, const QString &eventId)
+    {
+        QVariant empty = QVariant::fromValue(QDBusVariant(QString()));
+        uint timestamp = QDateTime::currentDateTime().toTime_t();
+        m_interface->asyncCall("Event", id, eventId, empty, timestamp);
+    }
 };
 
 DBusMenuImporter::DBusMenuImporter(const QString &service, const QString &path, QObject *parent)
@@ -333,22 +347,12 @@ QMenu *DBusMenuImporter::menu() const
     return d->m_menu;
 }
 
-void DBusMenuImporter::dispatch(QDBusPendingCallWatcher *watcher)
-{
-    Task task = d->m_taskForWatcher.take(watcher);
-    if (!task.m_method) {
-        DMWARNING << "No task for watcher!";
-        return;
-    }
-    (this->*task.m_method)(task.m_id, watcher);
-}
-
 void DBusMenuImporterPrivate::slotItemsPropertiesUpdated(const DBusMenuItemList &updatedList, const DBusMenuItemKeysList &removedList)
 {
     Q_FOREACH(const DBusMenuItem &item, updatedList) {
         QAction *action = m_actionForId.value(item.id);
         if (!action) {
-            DMWARNING << "No action for id" << item.id;
+            // We don't know this action. It probably is in a menu we haven't fetched yet.
             continue;
         }
 
@@ -363,7 +367,7 @@ void DBusMenuImporterPrivate::slotItemsPropertiesUpdated(const DBusMenuItemList
     Q_FOREACH(const DBusMenuItemKeys &item, removedList) {
         QAction *action = m_actionForId.value(item.id);
         if (!action) {
-            DMWARNING << "No action for id" << item.id;
+            // We don't know this action. It probably is in a menu we haven't fetched yet.
             continue;
         }
 
@@ -380,33 +384,10 @@ void DBusMenuImporter::slotItemActivationRequested(int id, uint /*timestamp*/)
     actionActivationRequested(action);
 }
 
-void DBusMenuImporter::GetPropertiesCallback(int id, QDBusPendingCallWatcher *watcher)
+void DBusMenuImporter::slotGetLayoutFinished(QDBusPendingCallWatcher *watcher)
 {
-    #ifdef BENCHMARK
-    DMDEBUG << "- Parsing updated properties for id" << id << sChrono.elapsed() << "ms";
-    #endif
-    QStringList requestedProperties = watcher->property("requestedProperties").toStringList();
-    QDBusReply<QVariantMap> reply = *watcher;
-    if (!reply.isValid()) {
-        DMWARNING << reply.error().message();
-        return;
-    }
+    int parentId = watcher->property(DBUSMENU_PROPERTY_ID).toInt();
 
-    QVariantMap properties = reply.value();
-
-    QAction *action = d->m_actionForId.value(id);
-    if (!action) {
-        DMWARNING << "No action for id" << id;
-        return;
-    }
-    d->updateAction(action, properties, requestedProperties);
-    #ifdef BENCHMARK
-    DMDEBUG << "- Item updated" << id << sChrono.elapsed() << "ms";
-    #endif
-}
-
-void DBusMenuImporter::GetLayoutCallback(int parentId, QDBusPendingCallWatcher *watcher)
-{
     QDBusPendingReply<uint, DBusMenuLayoutItem> reply = *watcher;
     if (!reply.isValid()) {
         DMWARNING << reply.error().message();
@@ -419,7 +400,10 @@ void DBusMenuImporter::GetLayoutCallback(int parentId, QDBusPendingCallWatcher *
     DBusMenuLayoutItem rootItem = reply.argumentAt<1>();
 
     QMenu *menu = d->menuForId(parentId);
-    DMRETURN_IF_FAIL(menu);
+    if (!menu) {
+        DMWARNING << "No menu for id" << parentId;
+        return;
+    }
 
     menu->clear();
 
@@ -445,9 +429,7 @@ void DBusMenuImporter::GetLayoutCallback(int parentId, QDBusPendingCallWatcher *
 
 void DBusMenuImporter::sendClickedEvent(int id)
 {
-    QVariant empty = QVariant::fromValue(QDBusVariant(QString()));
-    uint timestamp = QDateTime::currentDateTime().toTime_t();
-    d->m_interface->asyncCall("Event", id, QString("clicked"), empty, timestamp);
+    d->sendEvent(id, QString("clicked"));
 }
 
 void DBusMenuImporter::updateMenu()
@@ -525,6 +507,8 @@ void DBusMenuImporter::slotMenuAboutToShow()
     if (menu == d->m_menu) {
         menuReadyToBeShown();
     }
+
+    d->sendEvent(id, QString("opened"));
 }
 
 void DBusMenuImporter::slotAboutToShowDBusCallFinished(QDBusPendingCallWatcher *watcher)
@@ -550,6 +534,18 @@ void DBusMenuImporter::slotAboutToShowDBusCallFinished(QDBusPendingCallWatcher *
     }
 }
 
+void DBusMenuImporter::slotMenuAboutToHide()
+{
+    QMenu *menu = qobject_cast<QMenu*>(sender());
+    Q_ASSERT(menu);
+
+    QAction *action = menu->menuAction();
+    Q_ASSERT(action);
+
+    int id = action->property(DBUSMENU_PROPERTY_ID).toInt();
+    d->sendEvent(id, QString("closed"));
+}
+
 QMenu *DBusMenuImporter::createMenu(QWidget *parent)
 {
     return new QMenu(parent);
